/* tslint:disable:no-use-before-declare */
"use strict";
/**
 * No Case Found error message
 */
exports.NoCaseFound = 'no-case-found';
/**
 * Match statement
 * {{
 * Match( obj,
 *  Case( c => c === 'something', c => `c.toto`), ...
 * );
 * }}
 * @param a object to test
 * @param cases sets of cases to check.
 * @returns {B}
 * @throws 'no-case-found error' in case of no items match with cases.
 */
function Match(a, ...cases) {
    const item = cases.find(c => c.test(a));
    if (!item) {
        throw new Error(exports.NoCaseFound);
    }
    return item.exec(a);
}
exports.Match = Match;
function Case(test, exec) {
    return { test, exec };
}
exports.Case = Case;
